"""Sphinx configuration for vcspull documentation."""

# flake8: noqa: E501
from __future__ import annotations

import inspect
import pathlib
import sys
import typing as t
from os.path import relpath

import vcspull

if t.TYPE_CHECKING:
    from sphinx.application import Sphinx

# Get the project root dir, which is the parent dir of this
cwd = pathlib.Path.cwd()
project_root = cwd.parent
src_root = project_root / "src"

sys.path.insert(0, str(src_root))
sys.path.insert(0, str(cwd / "_ext"))

# package data
about: dict[str, str] = {}
with (src_root / "vcspull" / "__about__.py").open() as fp:
    exec(fp.read(), about)

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
    "sphinx.ext.todo",
    "sphinx.ext.napoleon",
    "sphinx.ext.linkcode",
    "sphinxarg.ext",  # sphinx-argparse
    "sphinx_inline_tabs",
    "sphinx_copybutton",
    "sphinxext.opengraph",
    "sphinxext.rediraffe",
    "myst_parser",
    "linkify_issues",
]
myst_enable_extensions = [
    "colon_fence",
    "substitution",
    "replacements",
    "strikethrough",
    "linkify",
]

templates_path = ["_templates"]

source_suffix = {".rst": "restructuredtext", ".md": "markdown"}

master_doc = "index"

project = about["__title__"]
project_copyright = about["__copyright__"]

version = "{}".format(".".join(about["__version__"].split("."))[:2])
release = "{}".format(about["__version__"])

exclude_patterns = ["_build"]

pygments_style = "monokai"
pygments_dark_style = "monokai"

html_favicon = "_static/favicon.ico"
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]
html_extra_path = ["manifest.json"]
html_theme = "furo"
html_theme_path: list[str] = []
html_theme_options = {
    "light_logo": "img/vcspull.svg",
    "dark_logo": "img/vcspull-dark.svg",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": about["__github__"],
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
    "source_repository": f"{about['__github__']}/",
    "source_branch": "master",
    "source_directory": "docs/",
}
html_sidebars = {
    "**": [
        "sidebar/scroll-start.html",
        "sidebar/brand.html",
        "sidebar/search.html",
        "sidebar/navigation.html",
        "sidebar/projects.html",
        "sidebar/scroll-end.html",
    ],
}

# linkify_issues
issue_url_tpl = about["__github__"] + "/issues/{issue_id}"

# sphinx.ext.autodoc
autoclass_content = "both"
autodoc_member_order = "bysource"
toc_object_entries_show_parents = "hide"
autodoc_default_options = {
    "undoc-members": True,
    "members": True,
    "private-members": True,
    "show-inheritance": True,
    "member-order": "bysource",
}

# sphinx-autodoc-typehints
autodoc_typehints = "description"  # show type hints in doc body instead of signature
simplify_optional_unions = True

# sphinx.ext.napoleon
napoleon_google_docstring = True
napoleon_include_init_with_doc = True

# sphinx-copybutton
copybutton_prompt_text = (
    r">>> |\.\.\. |> |\$ |\# | In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
)
copybutton_prompt_is_regexp = True
copybutton_remove_prompts = True

# sphinxext-rediraffe
rediraffe_redirects = "redirects.txt"
rediraffe_branch = "master~1"

# sphinxext.opengraph
ogp_site_url = about["__docs__"]
ogp_image = "_static/img/icons/icon-192x192.png"
ogp_site_name = about["__title__"]

intersphinx_mapping = {
    "py": ("https://docs.python.org/", None),
    "libvcs": ("http://libvcs.git-pull.com/", None),
}


def linkcode_resolve(domain: str, info: dict[str, str]) -> None | str:
    """
    Determine the URL corresponding to Python object.

    Notes
    -----
    From https://github.com/numpy/numpy/blob/v1.15.1/doc/source/conf.py, 7c49cfa
    on Jul 31. License BSD-3. https://github.com/numpy/numpy/blob/v1.15.1/LICENSE.txt
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except Exception:  # noqa: PERF203
            return None

    # strip decorators, which would resolve to the source of the decorator
    # possibly an upstream bug in getsourcefile, bpo-1764286
    try:
        unwrap = inspect.unwrap
    except AttributeError:
        pass
    else:
        if callable(obj):
            obj = unwrap(obj)

    try:
        fn = inspect.getsourcefile(obj)
    except Exception:
        fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except Exception:
        lineno = None

    linespec = f"#L{lineno}-L{lineno + len(source) - 1}" if lineno else ""

    fn = relpath(fn, start=pathlib.Path(vcspull.__file__).parent)

    if "dev" in about["__version__"]:
        return "{}/blob/master/{}/{}/{}{}".format(
            about["__github__"],
            "src",
            about["__package_name__"],
            fn,
            linespec,
        )
    return "{}/blob/v{}/{}/{}/{}{}".format(
        about["__github__"],
        about["__version__"],
        "src",
        about["__package_name__"],
        fn,
        linespec,
    )


def remove_tabs_js(app: Sphinx, exc: Exception) -> None:
    """Fix for sphinx-inline-tabs#18."""
    if app.builder.format == "html" and not exc:
        tabs_js = pathlib.Path(app.builder.outdir) / "_static" / "tabs.js"
        tabs_js.unlink(missing_ok=True)


def setup(app: Sphinx) -> None:
    """Sphinx setup hook."""
    app.connect("build-finished", remove_tabs_js)



"""Configuration functionality for vcspull."""

from __future__ import annotations

import copy
import fnmatch
import logging
import os
import pathlib
import typing as t

from libvcs.sync.git import GitRemote

from vcspull.validator import is_valid_config

from . import exc
from ._internal.config_reader import ConfigReader
from .util import get_config_dir, update_dict

log = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from collections.abc import Callable

    from typing_extensions import TypeGuard

    from .types import ConfigDict, RawConfigDict


def expand_dir(
    dir_: pathlib.Path,
    cwd: pathlib.Path | Callable[[], pathlib.Path] = pathlib.Path.cwd,
) -> pathlib.Path:
    """Return path with environmental variables and tilde ~ expanded.

    Parameters
    ----------
    _dir : pathlib.Path
    cwd : pathlib.Path, optional
        current working dir (for deciphering relative _dir paths), defaults to
        :py:meth:`os.getcwd()`

    Returns
    -------
    pathlib.Path :
        Absolute directory path
    """
    dir_ = pathlib.Path(os.path.expandvars(str(dir_))).expanduser()
    if callable(cwd):
        cwd = cwd()

    if not dir_.is_absolute():
        dir_ = pathlib.Path(os.path.normpath(cwd / dir_))
        assert dir_ == pathlib.Path(cwd, dir_).resolve(strict=False)
    return dir_


def extract_repos(
    config: RawConfigDict,
    cwd: pathlib.Path | Callable[[], pathlib.Path] = pathlib.Path.cwd,
) -> list[ConfigDict]:
    """Return expanded configuration.

    end-user configuration permit inline configuration shortcuts, expand to
    identical format for parsing.

    Parameters
    ----------
    config : dict
        the repo config in :py:class:`dict` format.
    cwd : pathlib.Path
        current working dir (for deciphering relative paths)

    Returns
    -------
    list : List of normalized repository information
    """
    configs: list[ConfigDict] = []
    if callable(cwd):
        cwd = cwd()

    for directory, repos in config.items():
        assert isinstance(repos, dict)
        for repo, repo_data in repos.items():
            conf: dict[str, t.Any] = {}

            """
            repo_name: http://myrepo.com/repo.git

            to

            repo_name: { url: 'http://myrepo.com/repo.git' }

            also assures the repo is a :py:class:`dict`.
            """

            if isinstance(repo_data, str):
                conf["url"] = repo_data
            else:
                conf = update_dict(conf, repo_data)

            if "repo" in conf:
                if "url" not in conf:
                    conf["url"] = conf.pop("repo")
                else:
                    conf.pop("repo", None)

            if "name" not in conf:
                conf["name"] = repo

            if "path" not in conf:
                conf["path"] = expand_dir(
                    pathlib.Path(expand_dir(pathlib.Path(directory), cwd=cwd))
                    / conf["name"],
                    cwd,
                )

            if "remotes" in conf:
                assert isinstance(conf["remotes"], dict)
                for remote_name, url in conf["remotes"].items():
                    if isinstance(url, GitRemote):
                        continue
                    if isinstance(url, str):
                        conf["remotes"][remote_name] = GitRemote(
                            name=remote_name,
                            fetch_url=url,
                            push_url=url,
                        )
                    elif isinstance(url, dict):
                        assert "push_url" in url
                        assert "fetch_url" in url
                        conf["remotes"][remote_name] = GitRemote(
                            name=remote_name,
                            **url,
                        )

            def is_valid_config_dict(val: t.Any) -> TypeGuard[ConfigDict]:
                assert isinstance(val, dict)
                return True

            assert is_valid_config_dict(conf)

            configs.append(conf)

    return configs


def find_home_config_files(
    filetype: list[str] | None = None,
) -> list[pathlib.Path]:
    """Return configs of ``.vcspull.{yaml,json}`` in user's home directory."""
    if filetype is None:
        filetype = ["json", "yaml"]
    configs: list[pathlib.Path] = []

    yaml_config = pathlib.Path("~/.vcspull.yaml").expanduser()
    has_yaml_config = yaml_config.exists()
    json_config = pathlib.Path("~/.vcspull.json").expanduser()
    has_json_config = json_config.exists()

    if not has_yaml_config and not has_json_config:
        log.debug(
            "No config file found. Create a .vcspull.yaml or .vcspull.json"
            " in your $HOME directory. http://vcspull.git-pull.com for a"
            " quickstart.",
        )
    else:
        if sum(filter(None, [has_json_config, has_yaml_config])) > 1:
            raise exc.MultipleConfigWarning
        if has_yaml_config:
            configs.append(yaml_config)
        if has_json_config:
            configs.append(json_config)

    return configs


def find_config_files(
    path: list[pathlib.Path] | pathlib.Path | None = None,
    match: list[str] | str | None = None,
    filetype: t.Literal["json", "yaml", "*"]
    | list[t.Literal["json", "yaml", "*"]]
    | None = None,
    include_home: bool = False,
) -> list[pathlib.Path]:
    """Return repos from a directory and match. Not recursive.

    Parameters
    ----------
    path : list
        list of paths to search
    match : list
        list of globs to search against
    filetype: list
        of filetypes to search against
    include_home : bool
        Include home configuration files

    Raises
    ------
    LoadConfigRepoConflict :
        There are two configs that have same path and name with different repo urls.

    Returns
    -------
    list :
        list of absolute paths to config files.
    """
    if filetype is None:
        filetype = ["json", "yaml"]
    if match is None:
        match = ["*"]
    config_files = []
    if path is None:
        path = get_config_dir()

    if include_home is True:
        config_files.extend(find_home_config_files())

    if isinstance(path, list):
        for p in path:
            config_files.extend(find_config_files(p, match, filetype))
            return config_files
    else:
        path = path.expanduser()
        if isinstance(match, list):
            for m in match:
                config_files.extend(find_config_files(path, m, filetype))
        elif isinstance(filetype, list):
            for f in filetype:
                config_files.extend(find_config_files(path, match, f))
        else:
            match = f"{match}.{filetype}"
            config_files = list(path.glob(match))

    return config_files


def load_configs(
    files: list[pathlib.Path],
    cwd: pathlib.Path | Callable[[], pathlib.Path] = pathlib.Path.cwd,
) -> list[ConfigDict]:
    """Return repos from a list of files.

    Parameters
    ----------
    files : list
        paths to config file
    cwd : pathlib.Path
        current path (pass down for :func:`extract_repos`

    Returns
    -------
    list of dict :
        expanded config dict item

    Todo
    ----
    Validate scheme, check for duplicate destinations, VCS urls
    """
    repos: list[ConfigDict] = []
    if callable(cwd):
        cwd = cwd()

    for file in files:
        if isinstance(file, str):
            file = pathlib.Path(file)
        assert isinstance(file, pathlib.Path)
        conf = ConfigReader._from_file(file)
        assert is_valid_config(conf)
        newrepos = extract_repos(conf, cwd=cwd)

        if not repos:
            repos.extend(newrepos)
            continue

        dupes = detect_duplicate_repos(repos, newrepos)

        if len(dupes) > 0:
            msg = ("repos with same path + different VCS detected!", dupes)
            raise exc.VCSPullException(msg)
        repos.extend(newrepos)

    return repos


ConfigDictTuple = tuple["ConfigDict", "ConfigDict"]


def detect_duplicate_repos(
    config1: list[ConfigDict],
    config2: list[ConfigDict],
) -> list[ConfigDictTuple]:
    """Return duplicate repos dict if repo_dir same and vcs different.

    Parameters
    ----------
    config1 : list[ConfigDict]

    config2 : list[ConfigDict]

    Returns
    -------
    list[ConfigDictTuple]
        List of duplicate tuples
    """
    if not config1:
        return []

    dupes: list[ConfigDictTuple] = []

    repo_dirs = {
        pathlib.Path(repo["path"]).parent / repo["name"]: repo for repo in config1
    }
    repo_dirs_2 = {
        pathlib.Path(repo["path"]).parent / repo["name"]: repo for repo in config2
    }

    for repo_dir, repo in repo_dirs.items():
        if repo_dir in repo_dirs_2:
            dupes.append((repo, repo_dirs_2[repo_dir]))

    return dupes


def in_dir(
    config_dir: pathlib.Path | None = None,
    extensions: list[str] | None = None,
) -> list[str]:
    """Return a list of configs in ``config_dir``.

    Parameters
    ----------
    config_dir : str
        directory to search
    extensions : list
        filetypes to check (e.g. ``['.yaml', '.json']``).

    Returns
    -------
    list
    """
    if extensions is None:
        extensions = [".yml", ".yaml", ".json"]
    if config_dir is None:
        config_dir = get_config_dir()

    return [
        path.name
        for path in config_dir.iterdir()
        if is_config_file(path.name, extensions) and not path.name.startswith(".")
    ]


def filter_repos(
    config: list[ConfigDict],
    path: pathlib.Path | t.Literal["*"] | str | None = None,
    vcs_url: str | None = None,
    name: str | None = None,
) -> list[ConfigDict]:
    """Return a :py:obj:`list` list of repos from (expanded) config file.

    path, vcs_url and name all support fnmatch.

    Parameters
    ----------
    config : dict
        the expanded repo config in :py:class:`dict` format.
    path : str, Optional
        directory of checkout location, fnmatch pattern supported
    vcs_url : str, Optional
        url of vcs remote, fn match pattern supported
    name : str, Optional
        project name, fnmatch pattern supported

    Returns
    -------
    list :
        Repos
    """
    repo_list: list[ConfigDict] = []

    if path:
        repo_list.extend(
            [
                r
                for r in config
                if fnmatch.fnmatch(str(pathlib.Path(r["path"]).parent), str(path))
            ],
        )

    if vcs_url:
        repo_list.extend(
            r
            for r in config
            if fnmatch.fnmatch(str(r.get("url", r.get("repo"))), vcs_url)
        )

    if name:
        repo_list.extend(
            [r for r in config if fnmatch.fnmatch(str(r.get("name")), name)],
        )

    return repo_list


def is_config_file(
    filename: str,
    extensions: list[str] | str | None = None,
) -> bool:
    """Return True if file has a valid config file type.

    Parameters
    ----------
    filename : str
        filename to check (e.g. ``mysession.json``).
    extensions : list or str
        filetypes to check (e.g. ``['.yaml', '.json']``).

    Returns
    -------
    bool : True if is a valid config file type
    """
    if extensions is None:
        extensions = [".yml", ".yaml", ".json"]
    extensions = [extensions] if isinstance(extensions, str) else extensions
    return any(filename.endswith(e) for e in extensions)


def save_config_yaml(config_file_path: pathlib.Path, data: dict[t.Any, t.Any]) -> None:
    """Save configuration data to a YAML file.

    Parameters
    ----------
    config_file_path : pathlib.Path
        Path to the configuration file to write
    data : dict
        Configuration data to save
    """
    yaml_content = ConfigReader._dump(
        fmt="yaml",
        content=data,
        indent=2,
    )
    config_file_path.write_text(yaml_content, encoding="utf-8")


def canonicalize_workspace_path(
    label: str,
    *,
    cwd: pathlib.Path | None = None,
) -> pathlib.Path:
    """Convert a workspace root label to an absolute canonical path."""
    cwd = cwd or pathlib.Path.cwd()
    label_path = pathlib.Path(label)
    return expand_dir(label_path, cwd=cwd)


def workspace_root_label(
    workspace_path: pathlib.Path,
    *,
    cwd: pathlib.Path | None = None,
    home: pathlib.Path | None = None,
) -> str:
    """Create a normalized label for a workspace root path."""
    cwd = cwd or pathlib.Path.cwd()
    home = home or pathlib.Path.home()

    if workspace_path == cwd:
        return "./"

    try:
        relative_to_home = workspace_path.relative_to(home)
        label = f"~/{relative_to_home.as_posix()}"
    except ValueError:
        label = workspace_path.as_posix()

    if label != "./" and not label.endswith("/"):
        label += "/"

    return label


def normalize_workspace_roots(
    config_data: dict[str, t.Any],
    *,
    cwd: pathlib.Path | None = None,
    home: pathlib.Path | None = None,
) -> tuple[dict[str, t.Any], dict[pathlib.Path, str], list[str], int]:
    """Normalize workspace root labels and merge duplicate sections."""
    cwd = cwd or pathlib.Path.cwd()
    home = home or pathlib.Path.home()

    normalized: dict[str, t.Any] = {}
    path_to_label: dict[pathlib.Path, str] = {}
    conflicts: list[str] = []
    change_count = 0

    for label, value in config_data.items():
        canonical_path = canonicalize_workspace_path(label, cwd=cwd)
        normalized_label = path_to_label.get(canonical_path)

        if normalized_label is None:
            normalized_label = workspace_root_label(
                canonical_path,
                cwd=cwd,
                home=home,
            )
            path_to_label[canonical_path] = normalized_label

            if isinstance(value, dict):
                normalized[normalized_label] = copy.deepcopy(value)
            else:
                normalized[normalized_label] = value

            if normalized_label != label:
                change_count += 1
        else:
            change_count += 1
            existing_value = normalized.get(normalized_label)

            if isinstance(existing_value, dict) and isinstance(value, dict):
                for repo_name, repo_config in value.items():
                    if repo_name not in existing_value:
                        existing_value[repo_name] = copy.deepcopy(repo_config)
                        change_count += 1
                    elif existing_value[repo_name] != repo_config:
                        conflict_message = (
                            "Workspace root '{label}' contains conflicting definitions "
                            "for repository '{repo}'. Keeping the existing entry."
                        )
                        conflicts.append(
                            conflict_message.format(
                                label=normalized_label,
                                repo=repo_name,
                            )
                        )
            elif existing_value != value:
                conflict_message = (
                    "Workspace root '{label}' contains conflicting non-dictionary "
                    "values. Keeping the existing entry."
                )
                conflicts.append(conflict_message.format(label=normalized_label))

    return normalized, path_to_label, conflicts, change_count



"""Tests for vcspull configuration format."""

from __future__ import annotations

import typing as t

import pytest

from vcspull import config

if t.TYPE_CHECKING:
    import pathlib

    from vcspull.types import ConfigDict


class LoadYAMLFn(t.Protocol):
    """Typing for load_yaml pytest fixture."""

    def __call__(
        self,
        content: str,
        path: str = "randomdir",
        filename: str = "randomfilename.yaml",
    ) -> tuple[pathlib.Path, list[t.Any | pathlib.Path], list[ConfigDict]]:
        """Callable function type signature for load_yaml pytest fixture."""
        ...


@pytest.fixture
def load_yaml(tmp_path: pathlib.Path) -> LoadYAMLFn:
    """Return a yaml loading function that uses temporary directory path."""

    def fn(
        content: str,
        path: str = "randomdir",
        filename: str = "randomfilename.yaml",
    ) -> tuple[pathlib.Path, list[pathlib.Path], list[ConfigDict]]:
        """Return vcspull configurations and write out config to temp directory."""
        dir_ = tmp_path / path
        dir_.mkdir()
        config_ = dir_ / filename
        config_.write_text(content, encoding="utf-8")

        configs = config.find_config_files(path=dir_)
        repos = config.load_configs(configs, cwd=dir_)
        return dir_, configs, repos

    return fn


def test_simple_format(load_yaml: LoadYAMLFn) -> None:
    """Test simple configuration YAML file for vcspull."""
    path, _, repos = load_yaml(
        """
vcspull:
  libvcs: git+https://github.com/vcs-python/libvcs
   """,
    )

    assert len(repos) == 1
    repo = repos[0]

    assert path / "vcspull" == repo["path"].parent
    assert path / "vcspull" / "libvcs" == repo["path"]


def test_relative_dir(load_yaml: LoadYAMLFn) -> None:
    """Test configuration files for vcspull support relative directories."""
    path, _, repos = load_yaml(
        """
./relativedir:
  docutils: svn+http://svn.code.sf.net/p/docutils/code/trunk
   """,
    )

    config_files = config.find_config_files(path=path)
    repos = config.load_configs(config_files, path)

    assert len(repos) == 1
    repo = repos[0]

    assert path / "relativedir" == repo["path"].parent
    assert path / "relativedir" / "docutils" == repo["path"]



"""Tests for vcspull utilities."""

from __future__ import annotations

import typing as t

from vcspull.util import get_config_dir

if t.TYPE_CHECKING:
    import pathlib

    import pytest


def test_vcspull_configdir_env_var(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test retrieving config directory with VCSPULL_CONFIGDIR set."""
    monkeypatch.setenv("VCSPULL_CONFIGDIR", str(tmp_path))

    assert get_config_dir() == tmp_path


def test_vcspull_configdir_xdg_config_dir(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test retrieving config directory with XDG_CONFIG_HOME set."""
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    vcspull_dir = tmp_path / "vcspull"
    vcspull_dir.mkdir()

    assert get_config_dir() == vcspull_dir


def test_vcspull_configdir_no_xdg(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test retrieving config directory without XDG_CONFIG_HOME set."""
    monkeypatch.delenv("XDG_CONFIG_HOME")
    assert get_config_dir()



"""Tests for vcspull import command functionality."""

from __future__ import annotations

import contextlib
import logging
import subprocess
import typing as t

import pytest
import yaml

from vcspull.cli import cli
from vcspull.cli._import import get_git_origin_url, import_from_filesystem, import_repo
from vcspull.config import canonicalize_workspace_path, workspace_root_label

if t.TYPE_CHECKING:
    import pathlib

    from libvcs.pytest_plugin import CreateRepoPytestFixtureFn
    from typing_extensions import TypeAlias

    ExpectedOutput: TypeAlias = t.Optional[t.Union[str, list[str]]]


def setup_git_repo(
    path: pathlib.Path,
    remote_url: str | None,
    git_envvars: dict[str, str],
) -> None:
    """Set up a git repository."""
    path.mkdir(parents=True, exist_ok=True)
    subprocess.run(
        ["git", "init"],
        cwd=path,
        check=True,
        capture_output=True,
        env=git_envvars,
    )

    if remote_url:
        subprocess.run(
            ["git", "remote", "add", "origin", remote_url],
            cwd=path,
            check=True,
            capture_output=True,
            env=git_envvars,
        )


def clone_repo(
    remote_url: str,
    local_path: pathlib.Path,
    git_envvars: dict[str, str],
) -> None:
    """Clone a git repository."""
    subprocess.run(
        ["git", "clone", remote_url, str(local_path)],
        check=True,
        capture_output=True,
        env=git_envvars,
    )


# =============================================================================
# Test fixtures for single repo import
# =============================================================================


class ImportRepoFixture(t.NamedTuple):
    """Pytest fixture for vcspull import command (single repo mode)."""

    # pytest internal: used for naming test
    test_id: str

    # test parameters
    cli_args: list[str]
    initial_config: dict[str, t.Any] | None
    expected_config_contains: dict[str, t.Any]
    expected_in_output: ExpectedOutput = None
    expected_not_in_output: ExpectedOutput = None
    expected_log_level: str = "INFO"
    should_create_config: bool = False


IMPORT_REPO_FIXTURES: list[ImportRepoFixture] = [
    # Simple repo import with default workspace root
    ImportRepoFixture(
        test_id="simple-repo-default-root",
        cli_args=["import", "myproject", "git@github.com:user/myproject.git"],
        initial_config=None,
        should_create_config=True,
        expected_config_contains={
            "./": {
                "myproject": {"repo": "git@github.com:user/myproject.git"},
            },
        },
        expected_in_output="Successfully imported 'myproject'",
    ),
    # Import with custom workspace root
    ImportRepoFixture(
        test_id="custom-workspace-root",
        cli_args=[
            "import",
            "mylib",
            "https://github.com/org/mylib",
            "--workspace-root",
            "~/projects/libs",
        ],
        initial_config=None,
        should_create_config=True,
        expected_config_contains={
            "~/projects/libs/": {
                "mylib": {"repo": "https://github.com/org/mylib"},
            },
        },
        expected_in_output="Successfully imported 'mylib'",
    ),
    # Import to existing config under specific workspace root
    ImportRepoFixture(
        test_id="import-to-existing",
        cli_args=[
            "import",
            "project2",
            "git@github.com:user/project2.git",
            "--workspace-root",
            "~/work",
        ],
        initial_config={
            "~/work/": {
                "project1": {"repo": "git@github.com:user/project1.git"},
            },
        },
        expected_config_contains={
            "~/work/": {
                "project1": {"repo": "git@github.com:user/project1.git"},
                "project2": {"repo": "git@github.com:user/project2.git"},
            },
        },
        expected_in_output="Successfully imported 'project2'",
    ),
    # Duplicate repo detection
    ImportRepoFixture(
        test_id="duplicate-repo",
        cli_args=[
            "import",
            "existing",
            "git@github.com:other/existing.git",
            "--workspace-root",
            "~/code",
        ],
        initial_config={
            "~/code/": {
                "existing": {"repo": "git@github.com:user/existing.git"},
            },
        },
        expected_config_contains={
            "~/code/": {
                "existing": {"repo": "git@github.com:user/existing.git"},
            },
        },
        expected_in_output=[
            "Repository 'existing' already exists",
            "Current URL: git@github.com:user/existing.git",
        ],
        expected_log_level="WARNING",
    ),
    # Path inference
    ImportRepoFixture(
        test_id="path-inference",
        cli_args=[
            "import",
            "inferred",
            "git@github.com:user/inferred.git",
            "--path",
            "~/dev/projects/inferred",
        ],
        initial_config=None,
        should_create_config=True,
        expected_config_contains={
            "~/dev/projects/inferred/": {
                "inferred": {"repo": "git@github.com:user/inferred.git"},
            },
        },
        expected_in_output="Successfully imported 'inferred'",
    ),
]


class ScanExistingFixture(t.NamedTuple):
    """Fixture for scan behaviour when configuration already includes a repo."""

    test_id: str
    workspace_root_key_style: str
    scan_arg_style: str


SCAN_EXISTING_FIXTURES: list[ScanExistingFixture] = [
    ScanExistingFixture(
        test_id="tilde-no-slash_scan-tilde",
        workspace_root_key_style="tilde_no_slash",
        scan_arg_style="tilde",
    ),
    ScanExistingFixture(
        test_id="tilde-no-slash_scan-abs",
        workspace_root_key_style="tilde_no_slash",
        scan_arg_style="absolute",
    ),
    ScanExistingFixture(
        test_id="tilde-with-slash_scan-tilde",
        workspace_root_key_style="tilde_with_slash",
        scan_arg_style="tilde",
    ),
    ScanExistingFixture(
        test_id="tilde-with-slash_scan-abs",
        workspace_root_key_style="tilde_with_slash",
        scan_arg_style="absolute",
    ),
    ScanExistingFixture(
        test_id="absolute-no-slash_scan-abs",
        workspace_root_key_style="absolute_no_slash",
        scan_arg_style="absolute",
    ),
    ScanExistingFixture(
        test_id="absolute-with-slash_scan-abs",
        workspace_root_key_style="absolute_with_slash",
        scan_arg_style="absolute",
    ),
]


@pytest.mark.parametrize(
    list(ImportRepoFixture._fields),
    IMPORT_REPO_FIXTURES,
    ids=[test.test_id for test in IMPORT_REPO_FIXTURES],
)
def test_import_repo_cli(
    tmp_path: pathlib.Path,
    capsys: pytest.CaptureFixture[str],
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    test_id: str,
    cli_args: list[str],
    initial_config: dict[str, t.Any] | None,
    expected_config_contains: dict[str, t.Any],
    expected_in_output: ExpectedOutput,
    expected_not_in_output: ExpectedOutput,
    expected_log_level: str,
    should_create_config: bool,
) -> None:
    """Test vcspull import command through CLI (single repo mode)."""
    caplog.set_level(expected_log_level)

    # Set up config file path
    config_file = tmp_path / ".vcspull.yaml"

    # Create initial config if provided
    if initial_config:
        yaml_content = yaml.dump(initial_config, default_flow_style=False)
        config_file.write_text(yaml_content, encoding="utf-8")

    # Add config path to CLI args if not specified
    if "-c" not in cli_args and "--config" not in cli_args:
        cli_args = [*cli_args[:1], "-c", str(config_file), *cli_args[1:]]

    # Change to tmp directory
    monkeypatch.chdir(tmp_path)

    # Run CLI command
    with contextlib.suppress(SystemExit):
        cli(cli_args)

    # Capture output
    captured = capsys.readouterr()
    output = "".join([*caplog.messages, captured.out, captured.err])

    # Check expected output (strip ANSI codes for comparison)
    import re

    clean_output = re.sub(r"\x1b\[[0-9;]*m", "", output)  # Strip ANSI codes

    if expected_in_output is not None:
        if isinstance(expected_in_output, str):
            expected_in_output = [expected_in_output]
        for needle in expected_in_output:
            assert needle in clean_output, (
                f"Expected '{needle}' in output, got: {clean_output}"
            )

    if expected_not_in_output is not None:
        if isinstance(expected_not_in_output, str):
            expected_not_in_output = [expected_not_in_output]
        for needle in expected_not_in_output:
            assert needle not in clean_output, f"Unexpected '{needle}' in output"

    # Verify config file
    if should_create_config or initial_config:
        assert config_file.exists(), "Config file should exist"

        # Load and verify config
        with config_file.open() as f:
            config_data = yaml.safe_load(f)

        # Check expected config contents
        for key, value in expected_config_contains.items():
            assert key in config_data, f"Expected key '{key}' in config"
            if isinstance(value, dict):
                for subkey, subvalue in value.items():
                    assert subkey in config_data[key], (
                        f"Expected '{subkey}' in config['{key}']"
                    )
                    assert config_data[key][subkey] == subvalue, (
                        f"Config mismatch for {key}/{subkey}: "
                        f"expected {subvalue}, got {config_data[key][subkey]}"
                    )


# =============================================================================
# Test fixtures for filesystem scan import
# =============================================================================


class ImportScanFixture(t.NamedTuple):
    """Pytest fixture for vcspull import --scan command."""

    # pytest internal: used for naming test
    test_id: str

    # test parameters
    repo_setup: list[tuple[str, str, bool]]  # (name, subdir, has_remote)
    cli_args: list[str]
    initial_config: dict[str, t.Any] | None
    expected_config_contains: dict[str, t.Any] | None
    expected_in_output: ExpectedOutput = None
    expected_not_in_output: ExpectedOutput = None
    expected_log_level: str = "INFO"
    should_create_config: bool = False
    user_input: str | None = None  # For confirmation prompts


IMPORT_SCAN_FIXTURES: list[ImportScanFixture] = [
    # Single repository scan
    ImportScanFixture(
        test_id="single-repo-scan",
        repo_setup=[("myproject", "", True)],  # One repo with remote
        cli_args=["import", "--scan", ".", "-y"],
        initial_config=None,
        should_create_config=True,
        expected_config_contains={"has_repos": True},  # Will verify dynamically
        expected_in_output=[
            "Found 1 new repository to import:",
            "+ myproject",
            "Successfully updated",
        ],
    ),
    # Multiple repositories non-recursive
    ImportScanFixture(
        test_id="multiple-repos-non-recursive-scan",
        repo_setup=[
            ("repo1", "", True),
            ("repo2", "", True),
            ("nested", "subdir", True),  # Should be ignored without -r
        ],
        cli_args=["import", "--scan", ".", "-y"],
        initial_config=None,
        should_create_config=True,
        expected_config_contains={"has_repos": True},
        expected_in_output=[
            "Found 2 new repositories to import:",
            "+ repo1",
            "+ repo2",
            "Successfully updated",
        ],
        expected_not_in_output="nested",
    ),
    # Recursive scan
    ImportScanFixture(
        test_id="recursive-scan",
        repo_setup=[
            ("repo1", "", True),
            ("nested", "subdir", True),
        ],
        cli_args=["import", "--scan", ".", "-r", "-y"],
        initial_config=None,
        should_create_config=True,
        expected_config_contains={"has_repos": True},
        expected_in_output=[
            "Found 2 new repositories to import:",
            "+ repo1",
            "+ nested",
            "Successfully updated",
        ],
    ),
    # Custom workspace root override
    ImportScanFixture(
        test_id="custom-workspace-root-scan",
        repo_setup=[("myrepo", "", True)],
        cli_args=[
            "import",
            "--scan",
            ".",
            "--workspace-root",
            "~/custom/path",
            "-y",
        ],
        initial_config=None,
        should_create_config=True,
        expected_config_contains={
            "~/custom/path/": {"myrepo": {}},
        },  # Just check repo exists
        expected_in_output=[
            "Found 1 new repository to import:",
            "Successfully updated",
        ],
    ),
    # No repositories found
    ImportScanFixture(
        test_id="no-repos-scan",
        repo_setup=[],  # No repositories
        cli_args=["import", "--scan", ".", "-y"],
        initial_config=None,
        should_create_config=False,
        expected_config_contains=None,
        expected_in_output="No git repositories found",
    ),
    # Repository without remote
    ImportScanFixture(
        test_id="repo-without-remote-scan",
        repo_setup=[("local_only", "", False)],  # No remote
        cli_args=["import", "--scan", ".", "-y"],
        initial_config=None,
        should_create_config=False,
        expected_config_contains=None,
        expected_in_output="No git repositories found",
        expected_log_level="WARNING",
    ),
    # All repositories already exist
    ImportScanFixture(
        test_id="all-existing-scan",
        repo_setup=[("existing1", "", True), ("existing2", "", True)],
        cli_args=["import", "--scan", ".", "-y"],
        initial_config={"dynamic": "will_be_set_in_test"},  # Will be set dynamically
        should_create_config=False,
        expected_config_contains=None,
        expected_in_output=[
            "Found 2 existing repositories",
            "All found repositories already exist",
        ],
    ),
    # Mixed existing and new
    ImportScanFixture(
        test_id="mixed-existing-new-scan",
        repo_setup=[
            ("existing", "", True),
            ("newrepo", "", True),
        ],
        cli_args=["import", "--scan", ".", "-y"],
        initial_config={"dynamic": "will_be_set_in_test"},  # Will be set for existing
        should_create_config=False,
        expected_config_contains={"has_repos": True},
        expected_in_output=[
            "Found 1 existing repositories",  # Note: plural form in message
            "Found 1 new repository to import:",
            "+ newrepo",
            "Successfully updated",
        ],
    ),
    # User confirmation - yes
    ImportScanFixture(
        test_id="user-confirm-yes-scan",
        repo_setup=[("repo_confirm", "", True)],
        cli_args=["import", "--scan", "."],  # No -y flag
        initial_config=None,
        should_create_config=True,
        expected_config_contains={"has_repos": True},
        expected_in_output=[
            "Found 1 new repository to import:",
            "Successfully updated",
        ],
        user_input="y\n",
    ),
    # User confirmation - no
    ImportScanFixture(
        test_id="user-confirm-no-scan",
        repo_setup=[("repo_no_confirm", "", True)],
        cli_args=["import", "--scan", "."],  # No -y flag
        initial_config=None,
        should_create_config=False,
        expected_config_contains=None,
        expected_in_output=[
            "Found 1 new repository to import:",
            "Aborted by user",
        ],
        user_input="n\n",
    ),
]


@pytest.mark.parametrize(
    list(ImportScanFixture._fields),
    IMPORT_SCAN_FIXTURES,
    ids=[test.test_id for test in IMPORT_SCAN_FIXTURES],
)
def test_import_scan_cli(
    tmp_path: pathlib.Path,
    capsys: pytest.CaptureFixture[str],
    caplog: pytest.LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
    create_git_remote_repo: CreateRepoPytestFixtureFn,
    git_commit_envvars: dict[str, str],
    test_id: str,
    repo_setup: list[tuple[str, str, bool]],
    cli_args: list[str],
    initial_config: dict[str, t.Any] | None,
    expected_config_contains: dict[str, t.Any] | None,
    expected_in_output: ExpectedOutput,
    expected_not_in_output: ExpectedOutput,
    expected_log_level: str,
    should_create_config: bool,
    user_input: str | None,
) -> None:
    """Test vcspull import --scan command through CLI."""
    # Set up scan directory
    scan_dir = tmp_path / "scan_dir"
    scan_dir.mkdir()

    # Set up repositories based on fixture
    repo_urls = {}
    for repo_name, subdir, has_remote in repo_setup:
        repo_parent = scan_dir / subdir if subdir else scan_dir
        repo_parent.mkdir(exist_ok=True, parents=True)
        repo_path = repo_parent / repo_name

        if has_remote:
            # Create remote and clone
            remote_path = create_git_remote_repo()
            remote_url = f"file://{remote_path}"
            clone_repo(remote_url, repo_path, git_commit_envvars)
            repo_urls[repo_name] = remote_url
        else:
            # Create local repo without remote
            setup_git_repo(repo_path, None, git_commit_envvars)

    # Set up config file
    config_file = tmp_path / ".vcspull.yaml"

    # Handle dynamic initial config for existing repo tests
    if initial_config and "dynamic" in initial_config:
        if test_id == "all-existing-scan":
            # All repos should be in config
            initial_config = {
                str(scan_dir) + "/": {
                    name: {"repo": repo_urls[name]}
                    for name, _, has_remote in repo_setup
                    if has_remote
                },
            }
        elif test_id == "mixed-existing-new-scan":
            # Only "existing" repo should be in config
            initial_config = {
                str(scan_dir) + "/": {"existing": {"repo": repo_urls["existing"]}},
            }

    if initial_config:
        yaml_content = yaml.dump(initial_config, default_flow_style=False)
        config_file.write_text(yaml_content, encoding="utf-8")

    # Update CLI args: replace "." with scan_dir and add config
    updated_cli_args = []
    for i, arg in enumerate(cli_args):
        if arg == "." and i > 0 and cli_args[i - 1] == "--scan":
            updated_cli_args.append(str(scan_dir))
        else:
            updated_cli_args.append(arg)

    # Insert config argument after "import"
    import_idx = updated_cli_args.index("import")
    updated_cli_args = [
        *updated_cli_args[: import_idx + 1],
        "-c",
        str(config_file),
        *updated_cli_args[import_idx + 1 :],
    ]

    # Change to tmp directory
    monkeypatch.chdir(tmp_path)

    # Mock user input if needed
    if user_input:
        monkeypatch.setattr("builtins.input", lambda _: user_input.strip())

    # Run CLI command
    with contextlib.suppress(SystemExit):
        cli(updated_cli_args)

    # Capture output
    captured = capsys.readouterr()
    output = "".join([*caplog.messages, captured.out, captured.err])

    # Strip ANSI codes for comparison
    import re

    clean_output = re.sub(r"\x1b\[[0-9;]*m", "", output)

    # Check expected output
    if expected_in_output is not None:
        if isinstance(expected_in_output, str):
            expected_in_output = [expected_in_output]
        for needle in expected_in_output:
            assert needle in clean_output, (
                f"Expected '{needle}' in output, got: {clean_output}"
            )

    if expected_not_in_output is not None:
        if isinstance(expected_not_in_output, str):
            expected_not_in_output = [expected_not_in_output]
        for needle in expected_not_in_output:
            assert needle not in clean_output, f"Unexpected '{needle}' in output"

    # Verify config file
    if should_create_config or (initial_config and expected_config_contains):
        assert config_file.exists(), "Config file should exist"

        # Load and verify config
        with config_file.open() as f:
            config_data = yaml.safe_load(f)

        # Check expected config contents
        if expected_config_contains:
            if "has_repos" in expected_config_contains:
                # Just check that repos were added
                assert config_data, "Config should have content"
                assert any(isinstance(v, dict) for v in config_data.values()), (
                    "Should have repo entries"
                )
            else:
                for key, value in expected_config_contains.items():
                    assert key in config_data, f"Expected key '{key}' in config"
                    if isinstance(value, dict):
                        for subkey, subvalue in value.items():
                            assert subkey in config_data[key], (
                                f"Expected '{subkey}' in config['{key}']"
                            )
                            # If subvalue is empty dict, just check that the key exists
                            if subvalue == {}:
                                assert isinstance(config_data[key][subkey], dict)
                            elif subvalue != t.Any:
                                assert config_data[key][subkey] == subvalue


# =============================================================================
# Unit tests
# =============================================================================


class TestImportRepoUnit:
    """Unit tests for import_repo function."""

    def test_import_repo_direct_call(
        self,
        tmp_path: pathlib.Path,
        caplog: pytest.LogCaptureFixture,
    ) -> None:
        """Test direct import_repo function call."""
        config_file = tmp_path / ".vcspull.yaml"

        # Call import_repo directly
        import_repo(
            name="direct-test",
            url="git@github.com:user/direct.git",
            config_file_path_str=str(config_file),
            path=None,
            workspace_root_path=None,
        )

        # Verify
        assert config_file.exists()
        with config_file.open() as f:
            config_data = yaml.safe_load(f)

        assert "./" in config_data
        assert "direct-test" in config_data["./"]
        assert config_data["./"]["direct-test"] == {
            "repo": "git@github.com:user/direct.git",
        }

    def test_import_repo_invalid_config(
        self,
        tmp_path: pathlib.Path,
        caplog: pytest.LogCaptureFixture,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test handling of invalid config file."""
        config_file = tmp_path / ".vcspull.yaml"

        # Write invalid YAML
        config_file.write_text("invalid: yaml: content:", encoding="utf-8")

        # Change to tmp directory
        monkeypatch.chdir(tmp_path)

        # Try to import repo and capture log output
        with caplog.at_level(logging.ERROR):
            import_repo(
                name="test",
                url="git@github.com:user/test.git",
                config_file_path_str=str(config_file),
                path=None,
                workspace_root_path=None,
            )

        assert "Error loading YAML" in caplog.text


class TestGetGitOriginUrl:
    """Unit tests for get_git_origin_url function."""

    def test_get_origin_url_success(
        self,
        create_git_remote_repo: CreateRepoPytestFixtureFn,
        tmp_path: pathlib.Path,
        git_commit_envvars: dict[str, str],
    ) -> None:
        """Test successfully getting origin URL."""
        # Create and clone a repo
        remote_path = create_git_remote_repo()
        remote_url = f"file://{remote_path}"
        local_path = tmp_path / "test_repo"

        clone_repo(remote_url, local_path, git_commit_envvars)

        # Test getting URL
        url = get_git_origin_url(local_path)
        assert url == remote_url

    def test_get_origin_url_no_remote(
        self,
        tmp_path: pathlib.Path,
        git_commit_envvars: dict[str, str],
    ) -> None:
        """Test handling repo without origin."""
        repo_path = tmp_path / "local_only"
        setup_git_repo(repo_path, None, git_commit_envvars)

        url = get_git_origin_url(repo_path)
        assert url is None

    def test_get_origin_url_not_git(
        self,
        tmp_path: pathlib.Path,
    ) -> None:
        """Test handling non-git directory."""
        regular_dir = tmp_path / "not_git"
        regular_dir.mkdir()

        url = get_git_origin_url(regular_dir)
        assert url is None


class TestImportFromFilesystemUnit:
    """Unit tests for import_from_filesystem function."""

    def test_import_scan_direct_call(
        self,
        create_git_remote_repo: CreateRepoPytestFixtureFn,
        tmp_path: pathlib.Path,
        git_commit_envvars: dict[str, str],
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        """Test direct import_from_filesystem call."""
        # Set up a repo
        scan_dir = tmp_path / "repos"
        scan_dir.mkdir()

        remote_path = create_git_remote_repo()
        remote_url = f"file://{remote_path}"
        repo_path = scan_dir / "test_repo"
        clone_repo(remote_url, repo_path, git_commit_envvars)

        config_file = tmp_path / ".vcspull.yaml"

        # Call function directly
        import_from_filesystem(
            scan_dir_str=str(scan_dir),
            config_file_path_str=str(config_file),
            recursive=False,
            workspace_root_override=None,
            yes=True,
        )

        # Verify config created
        assert config_file.exists()
        with config_file.open() as f:
            config_data = yaml.safe_load(f)

        expected_key = str(scan_dir) + "/"
        assert expected_key in config_data
        assert "test_repo" in config_data[expected_key]

    def test_many_existing_repos_summary(
        self,
        create_git_remote_repo: CreateRepoPytestFixtureFn,
        tmp_path: pathlib.Path,
        git_commit_envvars: dict[str, str],
        capsys: pytest.CaptureFixture[str],
        caplog: pytest.LogCaptureFixture,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        """Test summary output when many repos already exist."""
        scan_dir = tmp_path / "many_repos"
        scan_dir.mkdir()

        # Create many repos (>5 for summary mode)
        repo_data = {}
        for i in range(8):
            remote_path = create_git_remote_repo()
            remote_url = f"file://{remote_path}"
            repo_name = f"repo{i}"
            repo_path = scan_dir / repo_name
            clone_repo(remote_url, repo_path, git_commit_envvars)
            repo_data[repo_name] = {"repo": remote_url}

        # Pre-create config with all repos
        config_file = tmp_path / ".vcspull.yaml"
        initial_config = {str(scan_dir) + "/": repo_data}
        yaml_content = yaml.dump(initial_config, default_flow_style=False)
        config_file.write_text(yaml_content, encoding="utf-8")

        # Change to tmp directory
        monkeypatch.chdir(tmp_path)

        # Run scan through CLI
        with contextlib.suppress(SystemExit):
            cli(["import", "--scan", str(scan_dir), "-c", str(config_file), "-y"])

        # Check for summary message (not detailed list)
        captured = capsys.readouterr()
        output = "\n".join(caplog.messages) + captured.out + captured.err

        # Strip ANSI codes
        import re

        clean_output = re.sub(r"\x1b\[[0-9;]*m", "", output)

        assert "Found 8 existing repositories already in configuration" in clean_output
        assert "All found repositories already exist" in clean_output

    @pytest.mark.parametrize(
        list(ScanExistingFixture._fields),
        [
            (
                fixture.test_id,
                fixture.workspace_root_key_style,
                fixture.scan_arg_style,
            )
            for fixture in SCAN_EXISTING_FIXTURES
        ],
        ids=[fixture.test_id for fixture in SCAN_EXISTING_FIXTURES],
    )
    def test_scan_respects_existing_config_sections(
        self,
        tmp_path: pathlib.Path,
        monkeypatch: pytest.MonkeyPatch,
        test_id: str,
        workspace_root_key_style: str,
        scan_arg_style: str,
    ) -> None:
        """Ensure filesystem scan does not duplicate repositories in config."""
        home_dir = tmp_path / "home"
        scan_dir = home_dir / "study" / "c"
        repo_dir = scan_dir / "cpython"
        repo_git_dir = repo_dir / ".git"
        repo_git_dir.mkdir(parents=True, exist_ok=True)

        expected_repo_url = "git+https://github.com/python/cpython.git"
        config_file = home_dir / ".vcspull.yaml"

        if workspace_root_key_style == "tilde_no_slash":
            workspace_root_key = "~/study/c"
        elif workspace_root_key_style == "tilde_with_slash":
            workspace_root_key = "~/study/c/"
        elif workspace_root_key_style == "absolute_no_slash":
            workspace_root_key = str(scan_dir)
        elif workspace_root_key_style == "absolute_with_slash":
            workspace_root_key = str(scan_dir) + "/"
        else:
            error_msg = (
                f"Unhandled workspace_root_key_style: {workspace_root_key_style}"
            )
            raise AssertionError(error_msg)

        config_file.write_text(
            (f"{workspace_root_key}:\n  cpython:\n    repo: {expected_repo_url}\n"),
            encoding="utf-8",
        )

        monkeypatch.setenv("HOME", str(home_dir))
        monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
        monkeypatch.setattr(
            "vcspull.cli._import.get_git_origin_url",
            lambda _path: expected_repo_url,
        )

        if scan_arg_style == "tilde":
            scan_arg = "~/study/c"
        elif scan_arg_style == "absolute":
            scan_arg = str(scan_dir)
        else:
            error_msg = f"Unhandled scan_arg_style: {scan_arg_style}"
            raise AssertionError(error_msg)

        import_from_filesystem(
            scan_dir_str=scan_arg,
            config_file_path_str=None,
            recursive=False,
            workspace_root_override=None,
            yes=True,
        )

        with config_file.open(encoding="utf-8") as f:
            config_data = yaml.safe_load(f)

        expected_path = canonicalize_workspace_path(
            workspace_root_key,
            cwd=home_dir,
        )
        expected_label = workspace_root_label(
            expected_path,
            cwd=home_dir,
            home=home_dir,
        )
        assert expected_label in config_data
        assert "cpython" in config_data[expected_label]
        assert config_data[expected_label]["cpython"]["repo"] == expected_repo_url
        assert len(config_data) == 1
        assert "~/study/c" not in config_data


# =============================================================================
# Help and output tests
# =============================================================================


def test_import_command_help(
    capsys: pytest.CaptureFixture[str],
) -> None:
    """Test import command help output."""
    with contextlib.suppress(SystemExit):
        cli(["import", "--help"])

    captured = capsys.readouterr()
    output = captured.out + captured.err

    # Check help content
    assert "Import a repository to the vcspull configuration file" in output
    assert "name" in output
    assert "url" in output
    assert "--path" in output
    assert "--workspace-root" in output
    assert "--scan" in output
    assert "--config" in output


def test_import_scan_stream_output(
    tmp_path: pathlib.Path,
    capsys: pytest.CaptureFixture[str],
    monkeypatch: pytest.MonkeyPatch,
    create_git_remote_repo: CreateRepoPytestFixtureFn,
    git_commit_envvars: dict[str, str],
) -> None:
    """Ensure the CLI prints repo summaries to the user-facing stream."""
    scan_dir = tmp_path / "scan"
    scan_dir.mkdir()

    repo_name = "sample"
    remote_path = create_git_remote_repo()
    remote_url = f"file://{remote_path}"
    clone_repo(remote_url, scan_dir / repo_name, git_commit_envvars)

    config_file = tmp_path / ".vcspull.yaml"

    monkeypatch.chdir(tmp_path)

    with contextlib.suppress(SystemExit):
        cli(
            [
                "import",
                "--scan",
                str(scan_dir),
                "--yes",
                "-c",
                str(config_file),
            ],
        )

    captured = capsys.readouterr()

    import re

    clean_output = re.sub(r"\x1b\[[0-9;]*m", "", captured.out + captured.err)

    assert "Found 1 new repository to import:" in clean_output
    assert repo_name in clean_output



"""CLI utilities for vcspull."""

from __future__ import annotations

import argparse
import logging
import pathlib
import textwrap
import typing as t
from typing import overload

from libvcs.__about__ import __version__ as libvcs_version

from vcspull.__about__ import __version__
from vcspull.log import setup_logger

from ._import import (
    create_import_subparser,
    import_from_filesystem,
    import_repo,
)
from .fmt import create_fmt_subparser, format_config_file
from .sync import create_sync_subparser, sync

log = logging.getLogger(__name__)

SYNC_DESCRIPTION = textwrap.dedent(
    """
    sync vcs repos

    examples:
      vcspull sync "*"
      vcspull sync "django-*"
      vcspull sync "django-*" flask
      vcspull sync -c ./myrepos.yaml "*"
      vcspull sync -c ./myrepos.yaml myproject
""",
).strip()


@overload
def create_parser(
    return_subparsers: t.Literal[True],
) -> tuple[argparse.ArgumentParser, t.Any]: ...


@overload
def create_parser(return_subparsers: t.Literal[False]) -> argparse.ArgumentParser: ...


def create_parser(
    return_subparsers: bool = False,
) -> argparse.ArgumentParser | tuple[argparse.ArgumentParser, t.Any]:
    """Create CLI argument parser for vcspull."""
    parser = argparse.ArgumentParser(
        prog="vcspull",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=SYNC_DESCRIPTION,
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=f"%(prog)s {__version__}, libvcs {libvcs_version}",
    )
    parser.add_argument(
        "--log-level",
        metavar="level",
        action="store",
        default="INFO",
        help="log level (debug, info, warning, error, critical)",
    )

    subparsers = parser.add_subparsers(dest="subparser_name")
    sync_parser = subparsers.add_parser(
        "sync",
        help="synchronize repos",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=SYNC_DESCRIPTION,
    )
    create_sync_subparser(sync_parser)

    import_parser = subparsers.add_parser(
        "import",
        help="import repository or scan filesystem for repositories",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Import a repository to the vcspull configuration file. "
        "Can import a single repository by name and URL, or scan a directory "
        "to discover and import multiple repositories.",
    )
    create_import_subparser(import_parser)

    fmt_parser = subparsers.add_parser(
        "fmt",
        help="format vcspull configuration files",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Format vcspull configuration files for consistency. "
        "Normalizes compact format to verbose format, standardizes on 'repo' key, "
        "and sorts directories and repositories alphabetically.",
    )
    create_fmt_subparser(fmt_parser)

    if return_subparsers:
        # Return all parsers needed by cli() function
        return parser, (sync_parser, import_parser, fmt_parser)
    return parser


def cli(_args: list[str] | None = None) -> None:
    """CLI entry point for vcspull."""
    parser, subparsers = create_parser(return_subparsers=True)
    sync_parser, _import_parser, _fmt_parser = subparsers
    args = parser.parse_args(_args)

    setup_logger(log=log, level=args.log_level.upper())

    if args.subparser_name is None:
        parser.print_help()
        return
    if args.subparser_name == "sync":
        sync(
            repo_patterns=args.repo_patterns,
            config=pathlib.Path(args.config) if args.config else None,
            exit_on_error=args.exit_on_error,
            parser=sync_parser,
        )
    elif args.subparser_name == "import":
        # Unified import command
        if args.scan_dir:
            # Filesystem scan mode
            import_from_filesystem(
                scan_dir_str=args.scan_dir,
                config_file_path_str=args.config,
                recursive=args.recursive,
                workspace_root_override=args.workspace_root_path,
                yes=args.yes,
            )
        elif args.name and args.url:
            # Manual import mode
            import_repo(
                name=args.name,
                url=args.url,
                config_file_path_str=args.config,
                path=args.path,
                workspace_root_path=args.workspace_root_path,
            )
        else:
            # Error: need either name+url or --scan
            log.error("Either provide NAME and URL, or use --scan DIR")
            parser.exit(status=2)
    elif args.subparser_name == "fmt":
        format_config_file(args.config, args.write, args.all)



"""Tests for vcspull fmt command."""

from __future__ import annotations

import logging
import pathlib
import typing as t

import pytest
import yaml

from vcspull.cli.fmt import format_config, format_config_file, normalize_repo_config
from vcspull.config import (
    canonicalize_workspace_path,
    normalize_workspace_roots,
    workspace_root_label,
)

if t.TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


class WorkspaceRootFixture(t.NamedTuple):
    """Fixture for workspace root normalization cases."""

    test_id: str
    config_factory: t.Callable[[pathlib.Path], dict[str, t.Any]]


WORKSPACE_ROOT_FIXTURES: list[WorkspaceRootFixture] = [
    WorkspaceRootFixture(
        test_id="tilde-mixed-trailing-slash",
        config_factory=lambda _base: {
            "~/study/c": {
                "cpython": {"repo": "git+https://github.com/python/cpython.git"},
            },
            "~/study/c/": {
                "tmux": {"repo": "git+https://github.com/tmux/tmux.git"},
            },
        },
    ),
    WorkspaceRootFixture(
        test_id="home-vs-absolute",
        config_factory=lambda base: {
            str(base / "study" / "c"): {
                "cpython": {"repo": "git+https://github.com/python/cpython.git"},
            },
            "~/study/c/": {
                "tmux": {"repo": "git+https://github.com/tmux/tmux.git"},
            },
        },
    ),
    WorkspaceRootFixture(
        test_id="relative-vs-tilde",
        config_factory=lambda _base: {
            "./study/c": {
                "cpython": {"repo": "git+https://github.com/python/cpython.git"},
            },
            "~/study/c/": {
                "tmux": {"repo": "git+https://github.com/tmux/tmux.git"},
            },
        },
    ),
]


@pytest.mark.parametrize(
    list(WorkspaceRootFixture._fields),
    [
        (
            fixture.test_id,
            fixture.config_factory,
        )
        for fixture in WORKSPACE_ROOT_FIXTURES
    ],
    ids=[fixture.test_id for fixture in WORKSPACE_ROOT_FIXTURES],
)
def test_workspace_root_normalization(
    test_id: str,
    config_factory: t.Callable[[pathlib.Path], dict[str, t.Any]],
) -> None:
    """Ensure format_config merges duplicate workspace roots."""
    home_dir = pathlib.Path.home()
    config = config_factory(home_dir)

    canonical_paths = {
        canonicalize_workspace_path(label, cwd=home_dir) for label in config
    }
    expected_labels = [
        workspace_root_label(path, cwd=home_dir, home=home_dir)
        for path in sorted(canonical_paths, key=lambda p: p.as_posix())
    ]

    normalized_config, _map, conflicts, merge_changes = normalize_workspace_roots(
        config,
        cwd=home_dir,
        home=home_dir,
    )
    assert conflicts == []
    assert sorted(normalized_config.keys()) == expected_labels
    formatted, _changes = format_config(normalized_config)
    assert sorted(formatted.keys()) == expected_labels
    assert merge_changes >= len(config) - len(canonical_paths)


def test_normalize_repo_config_compact_to_verbose() -> None:
    """Compact repository config should expand to verbose form."""
    compact = "git+https://github.com/user/repo.git"
    normalized = normalize_repo_config(compact)
    assert normalized == {"repo": compact}


def test_normalize_repo_config_url_to_repo() -> None:
    """Entries using url key should be converted to repo."""
    config_with_url = {"url": "git+https://github.com/user/repo.git"}
    normalized = normalize_repo_config(config_with_url)
    assert normalized == {"repo": "git+https://github.com/user/repo.git"}


def test_normalize_repo_config_already_verbose() -> None:
    """Verbose configs should remain unchanged."""
    config = {"repo": "git+https://github.com/user/repo.git"}
    normalized = normalize_repo_config(config)
    assert normalized == config


def test_normalize_repo_config_preserves_extras() -> None:
    """Extra fields should be preserved during normalization."""
    config = {
        "url": "git+https://github.com/user/repo.git",
        "remotes": {"upstream": "git+https://github.com/upstream/repo.git"},
        "shell_command_after": "ln -sf /foo /bar",
    }
    normalized = normalize_repo_config(config)
    assert normalized == {
        "repo": "git+https://github.com/user/repo.git",
        "remotes": {"upstream": "git+https://github.com/upstream/repo.git"},
        "shell_command_after": "ln -sf /foo /bar",
    }


def test_normalize_repo_config_both_url_and_repo() -> None:
    """When url and repo keys coexist, keep config unchanged."""
    config = {
        "url": "git+https://github.com/user/repo1.git",
        "repo": "git+https://github.com/user/repo2.git",
    }
    normalized = normalize_repo_config(config)
    assert normalized == config


def test_format_config_sorts_directories() -> None:
    """Workspace roots should be sorted alphabetically."""
    config = {
        "~/zzz/": {"repo1": "url1"},
        "~/aaa/": {"repo2": "url2"},
        "~/mmm/": {"repo3": "url3"},
    }
    formatted, changes = format_config(config)
    assert list(formatted.keys()) == ["~/aaa/", "~/mmm/", "~/zzz/"]
    assert changes > 0


def test_format_config_sorts_repositories() -> None:
    """Repositories within a workspace root should be sorted."""
    config = {
        "~/projects/": {
            "zebra": "url1",
            "alpha": "url2",
            "beta": "url3",
        },
    }
    formatted, changes = format_config(config)
    assert list(formatted["~/projects/"].keys()) == ["alpha", "beta", "zebra"]
    assert changes > 0


def test_format_config_converts_compact_entries() -> None:
    """Compact repository entries should convert to verbose form."""
    config = {
        "~/projects/": {
            "repo1": "git+https://github.com/user/repo1.git",
            "repo2": {"url": "git+https://github.com/user/repo2.git"},
            "repo3": {"repo": "git+https://github.com/user/repo3.git"},
        },
    }
    formatted, changes = format_config(config)
    assert formatted["~/projects/"]["repo1"] == {
        "repo": "git+https://github.com/user/repo1.git",
    }
    assert formatted["~/projects/"]["repo2"] == {
        "repo": "git+https://github.com/user/repo2.git",
    }
    assert formatted["~/projects/"]["repo3"] == {
        "repo": "git+https://github.com/user/repo3.git",
    }
    assert changes == 2


def test_format_config_no_changes_when_already_normalized() -> None:
    """Formatter should report zero changes for already-normalized configs."""
    config = {
        "~/aaa/": {
            "alpha": {"repo": "url1"},
            "beta": {"repo": "url2"},
        },
        "~/bbb/": {
            "charlie": {"repo": "url3"},
        },
    }
    formatted, changes = format_config(config)
    assert formatted == config
    assert changes == 0


def test_format_config_complex_changes() -> None:
    """Formatter should handle sorting and conversions together."""
    config = {
        "~/zzz/": {
            "zebra": "compact-url",
            "alpha": {"url": "verbose-url"},
            "beta": {
                "repo": "already-good",
                "remotes": {"upstream": "upstream-url"},
            },
        },
        "~/aaa/": {
            "repo1": "another-compact",
        },
    }
    formatted, changes = format_config(config)
    assert list(formatted.keys()) == ["~/aaa/", "~/zzz/"]
    assert list(formatted["~/zzz/"].keys()) == ["alpha", "beta", "zebra"]
    assert formatted["~/aaa/"]["repo1"] == {"repo": "another-compact"}
    assert formatted["~/zzz/"]["zebra"] == {"repo": "compact-url"}
    assert formatted["~/zzz/"]["alpha"] == {"repo": "verbose-url"}
    assert formatted["~/zzz/"]["beta"]["repo"] == "already-good"
    assert formatted["~/zzz/"]["beta"]["remotes"] == {"upstream": "upstream-url"}
    assert changes > 0


def test_format_config_file_without_write(
    tmp_path: pathlib.Path,
    caplog: LogCaptureFixture,
) -> None:
    """format_config_file should log issues without modifying when write=False."""
    config_file = tmp_path / ".vcspull.yaml"
    original_config = {
        "~/zzz/": {
            "repo2": "url2",
            "repo1": {"url": "url1"},
        },
        "~/aaa/": {
            "repo3": "url3",
        },
    }

    config_file.write_text(yaml.dump(original_config), encoding="utf-8")

    with caplog.at_level(logging.INFO):
        format_config_file(str(config_file), write=False, format_all=False)

    saved_config = yaml.safe_load(config_file.read_text(encoding="utf-8"))
    assert saved_config == original_config
    assert "formatting issue" in caplog.text
    assert "Run with --write to apply" in caplog.text


def test_format_config_file_with_write(tmp_path: pathlib.Path) -> None:
    """format_config_file should rewrite file when write=True."""
    config_file = tmp_path / ".vcspull.yaml"
    original_config = {
        "~/zzz/": {
            "repo2": "url2",
            "repo1": {"url": "url1"},
        },
    }
    config_file.write_text(yaml.dump(original_config), encoding="utf-8")

    format_config_file(str(config_file), write=True, format_all=False)

    saved_config = yaml.safe_load(config_file.read_text(encoding="utf-8"))
    assert saved_config["~/zzz/"]["repo1"] == {"repo": "url1"}


def test_format_config_file_invalid_yaml(
    tmp_path: pathlib.Path,
    caplog: LogCaptureFixture,
) -> None:
    """Invalid YAML should be reported without crashing."""
    config_file = tmp_path / ".vcspull.yaml"
    config_file.write_text("invalid: yaml: content:", encoding="utf-8")

    with caplog.at_level(logging.ERROR):
        format_config_file(str(config_file), write=False, format_all=False)

    assert "Error loading config" in caplog.text


def test_format_config_file_missing_config(
    tmp_path: pathlib.Path,
    caplog: LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Formatting without available config should emit an error."""
    monkeypatch.chdir(tmp_path)

    with caplog.at_level(logging.ERROR):
        format_config_file(None, write=False, format_all=False)

    assert "No configuration file found" in caplog.text


def test_format_config_file_reports_changes(
    tmp_path: pathlib.Path,
    caplog: LogCaptureFixture,
) -> None:
    """Detailed change summary should be logged for pending updates."""
    config_file = tmp_path / ".vcspull.yaml"
    yaml_content = """~/zzz/:
  compact1: url1
  compact2: url2
  verbose1:
    url: url3
  verbose2:
    url: url4
~/aaa/:
  repo: url5
"""
    config_file.write_text(yaml_content, encoding="utf-8")

    with caplog.at_level(logging.INFO):
        format_config_file(str(config_file), write=False, format_all=False)

    text = caplog.text
    assert "compact to verbose format" in text
    assert "from 'url' to 'repo' key" in text
    assert "Directories will be sorted alphabetically" in text
    assert "Run with --write to apply" in text


def test_format_all_configs(
    tmp_path: pathlib.Path,
    caplog: LogCaptureFixture,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """format_config_file with --all should process discovered configs."""
    config_dir = tmp_path / ".config" / "vcspull"
    config_dir.mkdir(parents=True)

    home_config = tmp_path / ".vcspull.yaml"
    home_config.write_text(
        yaml.dump({"~/projects/": {"repo1": {"repo": "url1"}}}),
        encoding="utf-8",
    )

    work_config = config_dir / "work.yaml"
    work_config.write_text(
        """~/work/:
  repo2: url2
  repo1: url1
""",
        encoding="utf-8",
    )

    local_root = tmp_path / "project"
    local_root.mkdir()
    local_config = local_root / ".vcspull.yaml"
    local_config.write_text(
        yaml.dump({"./": {"repo3": {"url": "url3"}}}),
        encoding="utf-8",
    )

    monkeypatch.chdir(local_root)

    def fake_find_config_files(include_home: bool = False) -> list[pathlib.Path]:
        files: list[pathlib.Path] = [work_config]
        if include_home:
            files.insert(0, home_config)
        return files

    def fake_find_home_config_files(
        filetype: list[str] | None = None,
    ) -> list[pathlib.Path]:
        return [home_config]

    monkeypatch.setattr(
        "vcspull.cli.fmt.find_config_files",
        fake_find_config_files,
    )
    monkeypatch.setattr(
        "vcspull.cli.fmt.find_home_config_files",
        fake_find_home_config_files,
    )

    with caplog.at_level(logging.INFO):
        format_config_file(None, write=False, format_all=True)

    text = caplog.text
    assert "Found 3 configuration files to format" in text
    assert str(home_config) in text
    assert str(work_config) in text
    assert str(local_config) in text
    assert "already formatted correctly" in text
    assert "Repositories in ~/work/ will be sorted alphabetically" in text
    assert "All 3 configuration files processed successfully" in text



#!/usr/bin/env python
"""Example script for export gitlab organization to vcspull config file."""

from __future__ import annotations

import argparse
import logging
import os
import pathlib
import sys
import typing as t

import requests
import yaml
from libvcs.sync.git import GitRemote

from vcspull.cli.sync import CouldNotGuessVCSFromURL, guess_vcs

if t.TYPE_CHECKING:
    from vcspull.types import RawConfig

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format="%(message)s")

try:
    gitlab_token = os.environ["GITLAB_TOKEN"]
except KeyError:
    log.info("Please provide the environment variable GITLAB_TOKEN")
    sys.exit(1)

parser = argparse.ArgumentParser(
    description="Script to generate vcsconfig for all repositories \
    under the given namespace (needs Gitlab >= 10.3)",
)
parser.add_argument("gitlab_host", type=str, help="url to the gitlab instance")
parser.add_argument(
    "gitlab_namespace",
    type=str,
    help="namespace/group in gitlab to generate vcsconfig for",
)
parser.add_argument(
    "-c",
    type=str,
    help="path to the target config file (default: ./vcspull.yaml)",
    dest="config_file_name",
    required=False,
    default="./vcspull.yaml",
)

args = vars(parser.parse_args())
gitlab_host = args["gitlab_host"]
gitlab_namespace = args["gitlab_namespace"]
config_filename = pathlib.Path(args["config_file_name"])

try:
    if config_filename.is_file():
        result = input(
            f"The target config file ({config_filename}) already exists, \
            do you want to overwrite it? [y/N] ",
        )

        if result != "y":
            log.info(
                f"Aborting per user request as existing config file ({config_filename})"
                " should not be overwritten!",
            )
            sys.exit(0)

    config_file = config_filename.open(mode="w")
except OSError:
    log.info(f"File {config_filename} not accessible")
    sys.exit(1)

response = requests.get(
    f"{gitlab_host}/api/v4/groups/{gitlab_namespace}/projects",
    params={"include_subgroups": "true", "per_page": "100"},
    headers={"Authorization": f"Bearer {gitlab_token}"},
)

if response.status_code != 200:
    log.info(f"Error: {response}")
    sys.exit(1)

path_prefix = pathlib.Path().cwd()
config: RawConfig = {}


for group in response.json():
    url_to_repo = group["ssh_url_to_repo"].replace(":", "/")
    namespace_path = group["namespace"]["full_path"]
    reponame = group["path"]

    path = f"{path_prefix}/{namespace_path}"

    if path not in config:
        config[path] = {}

    # simplified config not working - https://github.com/vcs-python/vcspull/issues/332
    # config[path][reponame] = 'git+ssh://%s' % (url_to_repo)

    vcs = guess_vcs(url_to_repo)
    if vcs is None:
        raise CouldNotGuessVCSFromURL(url_to_repo)

    config[path][reponame] = {
        "name": reponame,
        "path": path / reponame,
        "url": f"git+ssh://{url_to_repo}",
        "remotes": {
            "origin": GitRemote(
                name="origin",
                fetch_url=f"ssh://{url_to_repo}",
                push_url=f"ssh://{url_to_repo}",
            ),
        },
        "vcs": vcs,
    }

config_yaml = yaml.dump(config)

log.info(config_yaml)

config_file.write(config_yaml)
config_file.close()
